#!/usr/bin/env python

#############################################
# rsgischkgdalfile.py
#
#  Copyright 2021 RSGISLib.
#
#  RSGISLib: 'The Remote Sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Check if a gdal file (raster or vector) is valid.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 22/12/2021
# Version: 2.0
# 
#############################################

import rsgislib.tools.checkdatasets
import argparse
import glob

if __name__ == "__main__":
    parser = argparse.ArgumentParser( description="A utility which can be used to check whether a GDAL "
                                                  "compatible file is valid and if there are any errors or warnings.")
    parser.add_argument("-i", "--input", type=str, required=True, help="Input file search path (i.e., '/path/to/*.kea')")
    parser.add_argument("--vec", action='store_true', default=False, help="Specify that the input file is a "
                                                                          "vector layer (otherwise assumed "
                                                                          "to be a raster).")
    parser.add_argument("--rmerr", action='store_true', default=False, help="Delete error files from system.")
    parser.add_argument("--printnames", action='store_true', default=False, help="Print file names as checking")
    parser.add_argument("--printerrs", action='store_true', default=False,
                        help="Print the error messages for the images")
    parser.add_argument("--nbands", type=int, default=0,
                        help="Check the number of bands is correct. Ignored if 0; Default.")
    parser.add_argument("--epsg", type=int, default=0, help="The EPSG code for the projection of the images.")
    parser.add_argument("--chkproj", action='store_true', default=False, help="Check that a projection is defined")
    parser.add_argument("--readimg", action='store_true', default=False, help="Check the image by reading part of it.")
    parser.add_argument("--chksum", action='store_true', default=False, help="Check the image by calculating a checksum from the image pixel values.")
    parser.add_argument("--vecmulti", action='store_true', default=False, help="Specifies that the vector file has multiple files (e.g., shapefile) and they should all be deleted if error and --rmerr is selected.")

    args = parser.parse_args()
    print(args.input)

    chk_projection = args.chkproj
    if (not args.chkproj) and (args.epsg > 0):
        chk_projection = True

    files = glob.glob(args.input)
    print("File Checks ({} Files Found):".format(len(files)))
    if args.vec:
        rsgislib.tools.checkdatasets.run_check_gdal_vector_files(files, chk_proj=chk_projection, epsg_code=args.epsg, max_file_size=None, rm_err=args.rmerr, print_err=True, multi_file=args.vecmulti, print_file_names=args.printnames)
    else:
        rsgislib.tools.checkdatasets.run_check_gdal_image_files(files, check_bands=True, n_bands=args.nbands, chk_proj=chk_projection, epsg_code=args.epsg, read_img=args.readimg, smpl_n_pxls=10, calc_chk_sum=args.chksum, rm_err=args.rmerr, print_err=args.printerrs, print_file_names=args.printnames)
    print("Finish Checks")




import pytest

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_get_regression_stats():
    import rsgislib.regression
    import numpy

    ref_data = numpy.random.rand(100)
    pred_data = numpy.random.rand(100)
    rsgislib.regression.get_regression_stats(ref_data, pred_data, n_vars=1)



import os
import pytest

H5PY_NOT_AVAIL = False
try:
    import h5py
except ImportError:
    H5PY_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
ZONALSTATS_DATA_DIR = os.path.join(DATA_DIR, "zonalstats")


def test_ext_point_band_values_file(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_pt_samples.geojson"
    )
    vec_lyr = "sen2_20210527_aber_pt_samples"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.ext_point_band_values_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        "testcolval",
        reproj_vec=False,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [39.0, 35.0, 124.0, 63.0, 36.0, 34.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_ext_point_band_values_file_reproj(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_utm30n.kea")
    vec_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_wgs84_pt_samples.geojson"
    )
    vec_lyr = "sen2_20210527_aber_wgs84_pt_samples"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.ext_point_band_values_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        "testcolval",
        reproj_vec=True,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [33.0, 188.0, 34.0, 26.0, 79.0, 67.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 1:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Min(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field="testcolval",
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [26.0, 38.0, 37.0, 163.0, 65.0, 41.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Max(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field="testcolval",
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [32.0, 48.0, 46.0, 417.0, 73.0, 52.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Mean(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field="testcolval",
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [
        29.299145299145298,
        43.858108108108105,
        41.67605633802817,
        309.48275862068965,
        69.29411764705883,
        46.58974358974359,
    ]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_StdDev(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field="testcolval",
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [
        1.1822400674899987,
        2.708869653402217,
        2.1609508072124695,
        68.0732975598059,
        1.8395877377778693,
        2.168751841414411,
    ]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Sum(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field="testcolval",
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [6856.0, 12982.0, 2959.0, 8975.0, 2356.0, 1817.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Count(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field="testcolval",
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [234.0, 296.0, 71.0, 29.0, 34.0, 39.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Mode(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field="testcolval",
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [29.0, 46.0, 40.0, 328.0, 69.0, 46.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Median(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=None,
        percentile_field=None,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field="testcolval",
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [29.0, 45.0, 42.0, 326.0, 69.0, 47.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_test_poly_pts_file_Percentile(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        percentile=75,
        percentile_field="testcolval",
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [30.0, 46.0, 43.0, 356.0, 70.0, 48.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_poly_pts_band_stats_file(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_poly_pts_band_stats_file(
        out_vec_file, out_vec_lyr, input_img, 1, "testcolval", vec_def_epsg=None
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [31.0, 46.0, 46.0, 326.0, 69.0, 49.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Min(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field="testcolval",
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [26.0, 38.0, 37.0, 163.0, 65.0, 41.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Max(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field="testcolval",
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [32.0, 48.0, 46.0, 417.0, 73.0, 52.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Mean(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field="testcolval",
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [
        29.299145299145298,
        43.858108108108105,
        41.67605633802817,
        309.48275862068965,
        69.29411764705883,
        46.58974358974359,
    ]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_StdDev(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field="testcolval",
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [
        1.1822400674899987,
        2.708869653402217,
        2.1609508072124695,
        68.0732975598059,
        1.8395877377778693,
        2.168751841414411,
    ]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Sum(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field="testcolval",
        count_field=None,
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [6856.0, 12982.0, 2959.0, 8975.0, 2356.0, 1817.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Count(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field="testcolval",
        mode_field=None,
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [234.0, 296.0, 71.0, 29.0, 34.0, 39.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Mode(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field="testcolval",
        median_field=None,
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [29.0, 46.0, 40.0, 328.0, 69.0, 46.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_calc_zonal_band_stats_file_Median(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    rsgislib.zonalstats.calc_zonal_band_stats_file(
        out_vec_file,
        out_vec_lyr,
        input_img,
        1,
        0,
        1000,
        0,
        min_field=None,
        max_field=None,
        mean_field=None,
        stddev_field=None,
        sum_field=None,
        count_field=None,
        mode_field=None,
        median_field="testcolval",
        vec_def_epsg=None,
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcolval")
    print(vals)
    ref_vals = [29.0, 45.0, 42.0, 326.0, 69.0, 47.0]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if abs(val - ref_val) > 0.0001:
            vals_eq = False
            break
    assert vals_eq


def test_image_zone_to_hdf(tmp_path):
    import rsgislib.zonalstats

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polygons.geojson")
    vec_lyr = "sen2_20210527_aber_polygons"

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.image_zone_to_hdf(
        input_img,
        vec_file,
        vec_lyr,
        out_h5_file,
        no_prj_warn=False,
        pxl_in_poly_method=rsgislib.zonalstats.METHOD_POLYCONTAINSPIXELCENTER,
    )

    assert os.path.exists(out_h5_file)


def test_extract_zone_img_values_to_hdf(tmp_path):
    import rsgislib.zonalstats

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polys.kea")

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_zone_img_values_to_hdf(
        input_img, in_msk_img, out_h5_file, 1, rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


def test_extract_zone_img_band_values_to_hdf(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_polys.kea")

    in_img_info = []
    in_img_info.append(
        rsgislib.imageutils.ImageBandInfo(input_img, "Image1", [1, 3, 4])
    )
    in_img_info.append(rsgislib.imageutils.ImageBandInfo(input_img, "Image2", [2]))

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_zone_img_band_values_to_hdf(
        in_img_info, in_msk_img, out_h5_file, 1, rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


def test_random_sample_hdf5_file(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")
    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.random_sample_hdf5_file(
        in_h5_file, out_h5_file, 250, 42, rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


def test_split_sample_hdf5_file(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")
    out_h5_p1_file = os.path.join(tmp_path, "out_h5_p1_file.h5")
    out_h5_p2_file = os.path.join(tmp_path, "out_h5_p2_file.h5")

    rsgislib.zonalstats.split_sample_hdf5_file(
        in_h5_file, out_h5_p1_file, out_h5_p2_file, 250, 42, rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_p1_file) and os.path.exists(out_h5_p2_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_merge_extracted_hdf5_data(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")
    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.merge_extracted_hdf5_data(
        [in_h5_file, in_h5_file], out_h5_file, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_get_var_from_hdf5_data():
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")

    data = rsgislib.zonalstats.get_var_from_hdf5_data(
        [in_h5_file, in_h5_file], var_idx=1
    )

    assert data.shape[0] == 1406


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_get_hdf5_data():
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")

    data = rsgislib.zonalstats.get_hdf5_data([in_h5_file, in_h5_file])

    assert (data.shape[0] == 1406) and (data.shape[1] == 6)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_msk_h5_smpls_to_finite_values(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")
    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.msk_h5_smpls_to_finite_values(
        in_h5_file,
        out_h5_file,
        datatype=rsgislib.TYPE_16INT,
        lower_limit=0,
        upper_limit=1000,
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_extract_chip_zone_image_band_values_to_hdf_no_rot(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_pts.kea")

    in_img_info = []
    in_img_info.append(
        rsgislib.imageutils.ImageBandInfo(input_img, "Image1", [1, 3, 4])
    )
    in_img_info.append(rsgislib.imageutils.ImageBandInfo(input_img, "Image2", [2]))

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf(
        in_img_info,
        in_msk_img,
        1,
        20,
        out_h5_file,
        rotate_chips=None,
        datatype=rsgislib.TYPE_16INT,
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_extract_chip_zone_image_band_values_to_hdf_with_rot(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_pts.kea")

    in_img_info = []
    in_img_info.append(
        rsgislib.imageutils.ImageBandInfo(input_img, "Image1", [1, 3, 4])
    )
    in_img_info.append(rsgislib.imageutils.ImageBandInfo(input_img, "Image2", [2]))

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf(
        in_img_info,
        in_msk_img,
        1,
        20,
        out_h5_file,
        rotate_chips=[30, 60],
        datatype=rsgislib.TYPE_16INT,
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_extract_ref_chip_zone_image_band_values_to_hdf_no_rot(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_pts.kea")

    in_img_info = []
    in_img_info.append(
        rsgislib.imageutils.ImageBandInfo(input_img, "Image1", [1, 3, 4])
    )
    in_img_info.append(rsgislib.imageutils.ImageBandInfo(input_img, "Image2", [2]))

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_ref_chip_zone_image_band_values_to_hdf(
        in_img_info,
        in_ref_img,
        1,
        in_msk_img,
        1,
        20,
        out_h5_file,
        rotate_chips=None,
        datatype=rsgislib.TYPE_16INT,
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_extract_ref_chip_zone_image_band_values_to_hdf_with_rot(tmp_path):
    import rsgislib.zonalstats
    import rsgislib.imageutils

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    in_msk_img = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_pts.kea")

    in_img_info = []
    in_img_info.append(
        rsgislib.imageutils.ImageBandInfo(input_img, "Image1", [1, 3, 4])
    )
    in_img_info.append(rsgislib.imageutils.ImageBandInfo(input_img, "Image2", [2]))

    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.extract_ref_chip_zone_image_band_values_to_hdf(
        in_img_info,
        in_ref_img,
        1,
        in_msk_img,
        1,
        20,
        out_h5_file,
        rotate_chips=[30, 60],
        datatype=rsgislib.TYPE_16INT,
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_split_sample_chip_hdf5_file(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_chip_vals.h5"
    )
    out_h5_p1_file = os.path.join(tmp_path, "out_h5_p1_file.h5")
    out_h5_p2_file = os.path.join(tmp_path, "out_h5_p2_file.h5")

    rsgislib.zonalstats.split_sample_chip_hdf5_file(
        in_h5_file, out_h5_p1_file, out_h5_p2_file, 3, 42, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_p1_file) and os.path.exists(out_h5_p2_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_merge_extracted_hdf5_chip_data(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_chip_vals.h5"
    )
    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.merge_extracted_hdf5_chip_data(
        [in_h5_file, in_h5_file], out_h5_file, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_split_sample_ref_chip_hdf5_file(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_refchip_vals.h5"
    )
    out_h5_p1_file = os.path.join(tmp_path, "out_h5_p1_file.h5")
    out_h5_p2_file = os.path.join(tmp_path, "out_h5_p2_file.h5")

    rsgislib.zonalstats.split_sample_ref_chip_hdf5_file(
        in_h5_file, out_h5_p1_file, out_h5_p2_file, 3, 42, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_p1_file) and os.path.exists(out_h5_p2_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_merge_extracted_hdf5_chip_ref_data(tmp_path):
    import rsgislib.zonalstats

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_refchip_vals.h5"
    )
    out_h5_file = os.path.join(tmp_path, "out_h5_file.h5")

    rsgislib.zonalstats.merge_extracted_hdf5_chip_ref_data(
        [in_h5_file, in_h5_file], out_h5_file, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)



#!/usr/bin/env python
"""
The tools.httptools
"""
from typing import Dict
import json
import os

import tqdm
import requests

import rsgislib
import rsgislib.tools.utils
import rsgislib.tools.filetools


class RSGISPyResponseException(rsgislib.RSGISPyException):
    def __init__(self, value, response=None):
        """
        Init for the RSGISPyResponseException class
        """
        self.value = value
        self.response = response

    def __str__(self):
        """
        Return a string representation of the exception
        """
        return "HTTP status {0} {1}: {2}".format(
            self.response.status_code, self.response.reason, repr(self.value)
        )


def check_http_response(response: requests.Response, url: str) -> bool:
    """
    Check the HTTP response and raise an exception with appropriate error message
    if request was not successful.

    :param response:
    :param url:
    :return:

    """
    try:
        response.raise_for_status()
        success = True
    except (requests.HTTPError, ValueError):
        success = False
        excpt_msg = "Invalid API response."
        try:
            excpt_msg = response.headers["cause-message"]
        except:
            try:
                excpt_msg = response.json()["error"]["message"]["value"]
            except:
                excpt_msg = (
                    "Unknown error ('{0}'), check url in a web browser: '{1}'".format(
                        response.reason, url
                    )
                )
        api_error = RSGISPyResponseException(excpt_msg, response)
        api_error.__cause__ = None
        raise api_error
    return success


def send_http_json_request(
    url: str,
    data: Dict = None,
    api_key: str = None,
    convert_to_json: bool = True,
    header_data: Dict = None,
) -> Dict:
    """
    A function which sends a http request with a json data packet.
    If an error occurs an exception will be raised.

    :param url: The URL for the request to be sent.
    :param data: dictionary of data which can be converted to str
                 using json.dumps.
    :param api_key: if provided then the api-key will be provided
                    via the http header.
    :param convert_to_json:
    :param header_data: a dict of header information.
    :return: A dict of data returned from the server.

    """
    if convert_to_json:
        if data is None:
            params_data = None
        else:
            params_data = json.dumps(data)
    else:
        params_data = data

    if api_key == None:
        response = requests.post(url, params_data, headers=header_data)
    else:
        if header_data is not None:
            header_data["X-Auth-Token"] = api_key
        else:
            header_data = {"X-Auth-Token": api_key}
        response = requests.post(url, params_data, headers=header_data)

    try:
        http_status_code = response.status_code
        if response == None:
            raise rsgislib.RSGISPyException("No output from service")

        if http_status_code == 404:
            raise rsgislib.RSGISPyException("404 Not Found")
        elif http_status_code == 401:
            raise rsgislib.RSGISPyException("401 Unauthorized")
        elif http_status_code == 400:
            raise rsgislib.RSGISPyException(f"Error Code: {http_status_code}")

        output = json.loads(response.text)
    except Exception as e:
        response.close()
        raise rsgislib.RSGISPyException(f"{e}")
    response.close()

    return output


def download_file_http(
    input_url: str,
    out_file_path: str,
    username: str = None,
    password: str = None,
    no_except: bool = True,
):
    """

    :param input_url:
    :param out_file_path:
    :param username:
    :param password:
    :return:

    """
    session_http = requests.Session()
    if (username is not None) and (password is not None):
        session_http.auth = (username, password)
    user_agent = "rsgislib/{}".format(rsgislib.get_rsgislib_version())
    session_http.headers["User-Agent"] = user_agent

    tmp_dwnld_path = out_file_path + ".incomplete"

    headers = {}

    try:
        with session_http.get(
            input_url, stream=True, auth=session_http.auth, headers=headers
        ) as r:
            if check_http_response(r, input_url):
                total = int(r.headers.get("content-length", 0))
                chunk_size = 2 ** 20
                n_chunks = int(total / chunk_size) + 1

                with open(tmp_dwnld_path, "wb") as f:
                    for chunk in tqdm.tqdm(
                        r.iter_content(chunk_size=chunk_size), total=n_chunks
                    ):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
        if os.path.exists(tmp_dwnld_path):
            os.rename(tmp_dwnld_path, out_file_path)
            print("Download Complete: {}".format(out_file_path))

    except Exception as e:
        if no_except:
            print(e)
        else:
            raise rsgislib.RSGISPyException("{}".format(e))
        return False
    return True


def wget_download_file(
    input_url: str,
    out_file_path: str,
    username: str = None,
    password: str = None,
    try_number: int = 10,
    time_out: int = 60,
    input_url_md5: str = None,
) -> (bool, str):
    """
    A function which downloads a file from a url using the wget command line tool.
    If a username or password are provided then both must be provided.

    :param input_url_md5:
    :param input_url: string with the URL to be downloaded.
    :param out_file_path: output file name and path.
    :param username: username for the download, if required. Default is None meaning
                     it will be ignored.
    :param password: password for the download, if required. Default is None meaning
                     it will be ignored.
    :param try_number: number of attempts at the download. Default is 10.
    :param time_out: number of seconds to time out Default is 60.
    :return: boolean specifying whether the file had been successfully downloaded
             and a string with user feedback (e.g., error message)

    """
    import subprocess

    try_number = str(try_number)
    time_out = str(time_out)
    success = False
    out_message = ""
    command = [
        "wget",
        "-c",
        "-O",
        out_file_path,
        "-t",
        f"{try_number}",
        "-T",
        f"{time_out}",
        "--no-check-certificate",
    ]
    if (username is not None) and (password is not None):
        command.append("--user")
        command.append(username)
        command.append("--password")
        command.append(password)
    command.append(input_url)

    download_state = -1
    try:
        download_state = subprocess.call(command)
    except Exception as e:
        out_message = (
            f"Download of file ({out_file_path}) failed.: Exception:\n" + e.__str__()
        )

    if download_state == 0:
        if input_url_md5 is not None:
            dwnld_file_md5 = rsgislib.tools.filetools.create_md5_hash(out_file_path)
            if dwnld_file_md5 == input_url_md5:
                success = True
                out_message = "File Downloaded and MD5 to checked."
            else:
                success = False
                out_message = "File Downloaded but MD5 did not match."
        else:
            success = True
            out_message = "File Downloaded but no MD5 to check against."

    if out_message == "":
        out_message = "File did not successfully download but no exception thrown."

    return success, out_message



#!/usr/bin/env python
"""
The tools.stats module contains some useful tools for calculating useful statistics which aren't
easily available else where.
"""
import numpy

import rsgislib


def calc_pandas_vif(df, cols=None):
    """
    A function to calculate variance inflation factors to
    investigate multicollinearity between predictor variables.

    Interpretation of VIF scores (somewhat subjective):
    1 = No multicollinearity.
    1-5 = Moderate multicollinearity.
    > 5 = High multicollinearity.
    > 10 = This predictor should be removed from the model.

    :param df: pandas dataframe where the columns are the predictor variables
    :param cols: list of columns in the dataframe
    :return: A pandas series containing the VIF for each predictor variable.

    .. code:: python

        df = pandas.read_csv('metrics.csv')
        cols = list(df.columns)
        # Subset to the column names of interest
        ind_vars = cols[6:]
        vifs_series = calc_pandas_vif(df, ind_vars)
        vifs_series.to_csv('VIF_scores.csv')

    """
    # Import the linear model module.
    import sklearn.linear_model
    import pandas

    # If the cols is None then get a list of all the columns
    if cols is None:
        cols = list(df.columns)

    # If there is less than 2 columns product error message.
    if len(cols) < 2:
        raise rsgislib.RSGISPyException(
            "The list of columns must be have a length of at least 2."
        )

    print("Calculating VIF for {} predictors variables...".format(len(cols)))

    # Create a linear model instance.
    lm = sklearn.linear_model.LinearRegression()

    # Create the dict for the output scores
    vif_scores = dict()
    # Iterative through the columns.
    for col in cols:
        # Get the list of columns without the current column
        tmp_cols = cols.copy()
        tmp_cols.remove(col)
        # Create numpy array for the y value to be fitted.
        y = df[col].values
        # Create numpy array for the x values
        x = df[tmp_cols].values
        lm.fit(x, y)
        # Use fitted model to predict y values
        y_pred = lm.predict(x)

        # Calculate the coefficient of determination:
        ss_tot = sum((y - numpy.mean(y)) ** 2)
        ss_res = sum((y - y_pred) ** 2)
        r2 = 1 - (ss_res / ss_tot)

        # Calculate VIF score:
        if r2 == 1:
            vif_scores[col] = 0.0
        else:
            vif_scores[col] = 1 / (1 - r2)

    # Create a pandas series and return it.
    return pandas.Series(vif_scores, name="VIF")


def cqv_threshold(df, cols=None, lowthreshold=0.25, highthreshold=0.75):
    """
    A function to remove features with low & high variance based on the coefficient
    of quartile variation (CQV).

    Low CQV == uninformative predictor.
    High CQV == unstable predictor variable.

    Inspired by sklearn.feature_selection.VarianceThreshold(), but more robust
    since the coefficient of quartile variation is independent of feature scaling.
    It is also less susceptible to outliers than the coefficient of variation.

    References:
    https://en.wikipedia.org/wiki/Quartile_coefficient_of_dispersion
    https://cran.r-project.org/web/packages/cvcqv/vignettes/cqv_versatile.html

    :param df: pandas dataframe where the columns are the predictor variables
    :param cols: list of columns in the dataframe
    :param lowthreshold: Float defining the CQV below which the predictors are uninfomative
    :param highthreshold: Float defining the CQV above which the predictors are unreliable
    :return: list of column names for good predictor variables

    """
    # If the cols is None then get a list of all the columns
    if cols is None:
        cols = list(df.columns)

    # If there is less than 2 columns product error message.
    if len(cols) < 2:
        raise rsgislib.RSGISPyException(
            "The list of columns must be have a length of at least 2."
        )

    # Create numpy array from the list columns
    x = df[cols].values

    # Calculate the Coefficient of Variation:
    print("Calculating CQV for {} predictor variables...".format(len(cols)))
    q1 = numpy.percentile(x, 25, axis=0)
    q3 = numpy.percentile(x, 75, axis=0)
    cqv = (q3 - q1) / (q3 + q1)

    print("Median CQV: {}".format(numpy.median(cqv)))

    # Index the good predictors:
    good_idx = numpy.where((cqv >= lowthreshold) & (cqv <= highthreshold))[0]

    # Get a list of column names
    good_cols = []
    for i in good_idx:
        good_cols.append(cols[i])

    print("Selected {} useful predictors...".format(len(good_cols)))

    return good_cols


def corr_feature_selection(df, dep_vars, ind_vars, n_min_clusters=3, n_max_clusters=25):
    """
    A function which performs a correlation based feature selection.
    This analysis will cluster the independent (predictor) variables based
    on the Pearson correlation distance metric. The Silhouette coefficient
    (Rousseeuw, 1987) is used to find the optimal number of clusters.

    :param df: pandas dataframe where the columns are the predictor variables
    :param dep_vars: list of dependent variables within the dataframe
    :param ind_vars: list of independent (predictor) variables within the dataframe
    :param n_min_clusters: The minimum number of clusters within the search (Default: 3)
    :param n_max_clusters: The maximum number of clusters within the search (Default: 25)
    :return: list of column names for good predictor variables

    """
    import tqdm
    import scipy.stats
    import sklearn.cluster
    import sklearn.metrics

    # Create a list for sorting the correlations for each dependent variable
    corr_data = list()
    # Loop through the dependent variables
    for dep_var in dep_vars:
        # Create a list for sorting the output correlation values
        corr_vals = list()
        # Loop through the independent variables
        for ind_var in ind_vars:
            # add the correlation values to the list
            corr_vals.append(df[dep_var].corr(df[ind_var], method="pearson"))
        # Add the list of correlation values to the overall list.
        corr_data.append(corr_vals)

    # Calculate the mean Pearson r value for each predictor variable,
    # using the Fisher z-transform to standardise the Pearson r values.
    # https://en.wikipedia.org/wiki/Fisher_transformation
    pearson_corr = numpy.tanh(numpy.mean(numpy.arctanh(corr_data), axis=0))

    def _cluster_features(X, n_clusters):
        """
        An internal function to cluster predictor variables using feature agglomeration.

        :params X: a 2D numpy array of shape = (nSamples, nPredictors).
        :params n_clusters: the number of clusters to output
        :returns: the labels
        """
        # check inputs:
        if X.ndim != 2:
            raise rsgislib.RSGISPyException("Error: the input array must be 2D.")

        def _pearson_affinity(a):
            """An utility function to generate an affinity matrix based on Pearson correlation."""
            return 1 - numpy.array(
                [[abs(scipy.stats.pearsonr(x, y)[0]) for x in a] for y in a]
            )

        clf = sklearn.cluster.FeatureAgglomeration(
            n_clusters=n_clusters, linkage="complete", affinity=_pearson_affinity
        )
        clf.fit(X)

        # Get the cluster id for each feature/predictor variable:
        labels = clf.labels_ + 1

        return labels

    # Create the nd array from the pandas data range with all the predictor
    # variables listed.
    x = df[ind_vars].values

    # Test for optimal number of clusters using the Silhouette Coefficient as a
    # performance metric.

    n_clusters = numpy.arange(n_min_clusters, n_max_clusters, dtype="uint8")
    cluster_score = list()
    for n in tqdm.tqdm(n_clusters):
        x_lbls = _cluster_features(x, n)
        cluster_score.append(
            sklearn.metrics.silhouette_score(x.T, x_lbls, metric="correlation")
        )

    # find the maximum Silhouette Coefficient = best clustering:
    best_idx = numpy.argmax(cluster_score)
    opt_n_clusters = n_clusters[best_idx]
    opt_n_clusters_scr = cluster_score[best_idx]
    print("Found optimal number of clusters: {}".format(opt_n_clusters))
    print("Silhouette Coefficient: {}".format(opt_n_clusters_scr))

    # Perform clustering with optimal number of clusters
    cluster_lbls = _cluster_features(x, opt_n_clusters)

    # Select one predictor variable from each cluster using the maximum Pearson correlation:
    best_predictors = []
    for lbl in numpy.unique(cluster_lbls):
        lbl_idx = numpy.where(cluster_lbls == lbl)[0]
        best_corr = numpy.argmax(pearson_corr[lbl_idx])
        best_predictors.append(ind_vars[lbl_idx[best_corr]])

    return best_predictors


def lassolars_feature_selection(df, dep_vars, ind_vars, alpha_val=None):
    """
    A function which undertake regularisation-based feature selection using
    the LassoLars regressor in Scikit-Learn. the Lasso (least absolute shrinkage
    and selection operator) regression algorithm is linear model that uses L1
    regularisation to assign coefficients of zero to uninformative predictor
    variables (effectively eliminating them from the regression model). The
    LARS algorithm (Efron et al., 2004) provides a means of estimating which
    variables to include in the model, as well as their coefficients.

    :param df: pandas dataframe where the columns are the predictor variables
    :param dep_vars: list of dependent variables within the dataframe
    :param ind_vars: list of independent (predictor) variables within the dataframe
    :param alpha_val: Value of the regularization parameter (alpha) for the Lasso estimator.
                      If None then the value will be estimated using Bayes Information criterion
                      (BIC) and cross-validation. (Default: None).
    :return: list of column names for good predictor variables

    """
    import warnings
    import sklearn.linear_model
    import sklearn.feature_selection
    import sklearn.exceptions

    # Create the nd array from the pandas data range with all the predictor
    # variables listed.
    x = df[ind_vars].values

    # Create the nd array from the pandas data range with all the dependent
    # variables listed.
    y = df[dep_vars].values

    if alpha_val is None:
        # Use the Bayes Information criterion (BIC) and cross-validation to find the
        # optimal value of the regularization parameter (alpha) for the Lasso estimator.
        estimator = sklearn.linear_model.LassoLarsIC("bic")

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", category=sklearn.exceptions.ConvergenceWarning
            )
            estimator.fit(x, y[:, -1])

        optimal_alpha = estimator.alpha_
    else:
        optimal_alpha = float(alpha_val)

    print(
        "Using regularization parameter (alpha) for the Lasso estimator of: {}".format(
            optimal_alpha
        )
    )

    # Fit the regressor using tuned regularisation parameter:
    estimator = sklearn.linear_model.LassoLars(alpha=optimal_alpha)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", category=sklearn.exceptions.ConvergenceWarning
        )
        estimator.fit(x, y)

    # Select the variables with non-zero coefficients:
    selector = sklearn.feature_selection.SelectFromModel(estimator, prefit=True)
    ind_vars_arr = numpy.array(ind_vars)
    best_predictors = ind_vars_arr[selector.get_support()]

    return list(best_predictors)


def breusch_pagan_test(x, y):
    """
    A function to perform a Breusch-Pagan test for heteroskedasticity in a linear model:
    H_0 = No heteroskedasticity.
    H_1 = Heteroskedasticity is present.

    Returns a list containing three elements:
    1. the Breusch-Pagan test statistic.
    2. the p-value for the test.
    3. the test result.

    :param x: a numpy.ndarray containing the predictor variables. Shape = (nSamples, nPredictors).
    :param y: a 1D numpy.ndarray containing the response variable. Shape = (nSamples, ).
    :return: list containing 3 elements (Breusch-Pagan test statistic, p-value for the test, test result)

    """
    import sklearn.linear_model

    if x.shape[0] != y.shape[0]:
        raise SystemExit("Error: the number of samples differs between x and y.")
    else:
        n_samples = x.shape[0]

    # fit an OLS linear model to y using x:
    lm = sklearn.linear_model.LinearRegression()
    lm.fit(x, y)

    # calculate the squared errors:
    err = (y - lm.predict(x)) ** 2

    # fit an auxiliary regression to the squared errors:
    # why?: to estimate the variance in err explained by x
    lm.fit(x, err)
    pred_err = lm.predict(x)

    # calculate the coefficient of determination:
    ss_tot = numpy.sum((err - numpy.mean(err, axis=0)) ** 2)
    ss_res = numpy.sum((err - pred_err) ** 2)
    r2 = 1 - (ss_res / ss_tot)

    # calculate the Lagrange multiplier:
    LM = n_samples * r2

    # calculate p-value. degrees of freedom = number of predictors.
    # this is equivalent to (p - 1) parameter restrictions in Wikipedia entry.
    try:
        from scipy.stats import chisqprob

        pval = chisqprob(LM, x.shape[1])
    except Exception:
        from scipy.stats.distributions import chi2

        pval = chi2.sf(LM, x.shape[1])

    if pval < 0.01:
        test_result = "Heteroskedasticity present at 99% CI."
    elif pval < 0.05:
        test_result = "Heteroskedasticity present at 95% CI."
    else:
        test_result = "No significant heteroskedasticity."
    return [LM, pval, test_result]


def bin_accuracy_scores_prob(y_true, y_prob):
    """
    A function to calculate accuracy measures for probabilistic responses with sklearn and scipy.
    Function written by Osian Roberts.

    :param y_true: binary class labels, where 0 is absence and 1 is presence.
    :param y_prob: probability of presence scores e.g., generated by a species distribution model.
    :returns: a list containing two arrays - metrics = names of test metrics. scores = test scores for each metric.

    Useful reference:
    https://machinelearningmastery.com/how-to-score-probability-predictions-in-python

    """
    # check inputs:
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_prob, numpy.ndarray):
        y_prob = numpy.array(y_prob)
    if y_true.ndim != 1:
        raise SystemExit("ERROR: the true labels are not in a 1D array.")
    if y_prob.ndim != 1:
        raise SystemExit(
            "ERROR: the probability of presence values are not in a 1D array."
        )
    if y_true.size != y_prob.size:
        raise SystemExit("ERROR: unequal number of binary labels and probabilities.")

        # ensure that y_true contains binary labels (i.e. 0 or 1 values):
    y_true = y_true.astype("uint8")
    if numpy.min(y_true) != 0 or numpy.max(y_true) != 1:
        raise SystemExit("ERROR: the true labels are not binary (zero or one values).")

    from sklearn.metrics import roc_auc_score

    # calculates area under the receiver operating curve score.
    # A score of 0.5 shows the model is unable to discriminate between presence and absence.
    roc_auc = roc_auc_score(y_true, y_prob)

    from sklearn.metrics import average_precision_score

    # calculates area under the precision-recall curve. Perfect model = 1.0.
    average_precision = average_precision_score(y_true, y_prob)

    from sklearn.metrics import brier_score_loss

    # This is a quadratic loss function that calculates the mean squared error between
    # predicted probabilities and the true presence-absence (binary) labels.
    # A model with no false positives/negatives has a score of 0.0. Perfect model = 1.0.
    brier_score = brier_score_loss(y_true, y_prob)

    from sklearn.metrics import log_loss

    # The is logarithmic loss function that more heavily penalises false positives/negatives than the brier score.
    # A model with no false positives/negatives has a score of 0.0. There is no upper bound.
    log_loss_score = log_loss(y_true, y_prob)

    from scipy.stats import pointbiserialr

    # The point biserial correlation coefficient, range -1 to 1.
    # Quantifies the correlation between a binary and continuous variable.
    r = pointbiserialr(y_true, y_prob)[0]

    metrics = [
        "Test AUC",
        "Point-Biserial r",
        "Av. Precision",
        "Brier Score",
        "Log-Loss Score",
    ]
    scores = numpy.array(
        [roc_auc, r, average_precision, brier_score, log_loss_score]
    ).round(decimals=6)
    del roc_auc, r, average_precision, brier_score, log_loss_score, y_true, y_prob
    return [metrics, scores]


def accuracy_scores_binary(y_true, y_pred):
    """
    A function to calculate accuracy measures for a binary classification.
    Function written by Osian Roberts.

    Parameters:
    :param y_true: observed binary labels, where 0 is absence and 1 is presence.
    :param y_pred: predicted binary labels, where 0 is absence and 1 is presence.
    :returns: a list containing two numpy.arrays - (metrics: name of test metrics, scores: test scores for each metric)

    Reference: See pages 253 - 255 in:
    Guisan et al. (2017). Habitat suitability and distribution models: with applications in R.

    """
    # check inputs:
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.array(y_pred)
    if y_true.ndim != 1:
        raise SystemExit("ERROR: the true labels are not in a 1D array.")
    if y_pred.ndim != 1:
        raise SystemExit("ERROR: the predicted labels are not in a 1D array.")
    if y_true.size != y_pred.size:
        raise SystemExit("ERROR: unequal number of binary labels.")

    # ensure that y_true, y_pred contain binary labels (i.e. 0 or 1 values):
    y_true = y_true.astype("uint8")
    y_pred = y_pred.astype("uint8")
    if numpy.min(y_true) != 0 or numpy.max(y_true) != 1:
        raise SystemExit("ERROR: the true labels are not binary (zero or one values).")
    if numpy.min(y_pred) != 0 or numpy.max(y_pred) != 1:
        raise SystemExit(
            "ERROR: the predicted labels are not binary (zero or one values)."
        )

    metrics = numpy.array(
        [
            "Prevalence",
            "Overall Diagnostic Power",
            "Correct Classification Rate",
            "Misclassification Rate",
            "Presence Predictive Power",
            "Absence Predictive Power",
            "Accuracy",
            "Balanced Accuracy",
            "Sensitivity",
            "Specificity",
            "Precision",
            "F1 Score",
            "Matthews Correlation",
            "Cohen Kappa",
            "Normalised Mutual Information",
            "Hanssen-Kuiper skill",
        ]
    )

    try:
        n_presence = numpy.where(y_true == 1)[0].size
        n_absence = numpy.where(y_true == 0)[0].size

        # calculate true-presence, true-absence, false-presence and false-absence:
        TP = numpy.where((y_true == 1) & (y_pred == 1))[0].size
        TA = numpy.where((y_true == 0) & (y_pred == 0))[0].size
        FP = numpy.where((y_true == 1) & (y_pred == 0))[0].size
        FA = numpy.where((y_true == 0) & (y_pred == 1))[0].size  # aka sweet FA!

        # proportion of presence records:
        prevalence = (TP / FA) / y_true.size

        # proportion of absence records:
        ODP = 1 - prevalence

        # correct classification & misclassification rate
        CCR = (TP + TA) / y_true.size
        MR = (FP + FA) / y_true.size

        # Sensitivity (aka Recall or True Positive Rate):
        sensitivity = TP / n_presence

        # false presence rate - inverse of sensitivity (redundant?)
        # FPR = 1  - sensitivity

        # Presence and absence predictive power:
        PPP = TP / (TP + FP)
        APP = TA / (TA + FA)

        # Specificity (aka True Negative Rate):
        specificity = TA / n_absence

        # false positive rate - inverse of specificity (redundant?)
        # FPR = 1 - specificity

        # Accuracy scores:
        accuracy = (TP + TA) / (n_presence + n_absence)
        balanced_accuracy = ((TP / n_presence) + (TA / n_absence)) / 2

        # precision:
        precision = TP / (TP + FP)

        # F1 score:
        f1_score = 2 * TP / ((2 * TP) + FP + FA)

        # Matthews Correlation Coefficient:
        MCC = ((TP * TA) - (FP * FA)) / (
            ((TP + FP) * (TP + FA) * (TA + FP) * (TA + FA)) ** 0.5
        )

        # Hanssen-Kuiper skill (unreliable when TA is very large):
        TSS = sensitivity + specificity - 1
        del n_presence, n_absence, TP, TA, FP, FA

        from sklearn.metrics import normalized_mutual_info_score as nmi_score

        nmi_score = nmi_score(y_true, y_pred)

        # Cohen's Kappa (caution: sensitive to sample size and proportion of presence records):
        from sklearn.metrics import cohen_kappa_score as kappa

        kappa = kappa(y_true, y_pred)

        scores = numpy.array(
            [
                prevalence,
                ODP,
                CCR,
                MR,
                PPP,
                APP,
                accuracy,
                balanced_accuracy,
                sensitivity,
                specificity,
                precision,
                f1_score,
                MCC,
                kappa,
                nmi_score,
                TSS,
            ]
        ).round(decimals=6)
        del prevalence, ODP, CCR, MR, PPP, APP, accuracy, balanced_accuracy, sensitivity
        del specificity, precision, f1_score, MCC, kappa, nmi_score, TSS
    except Exception:
        scores = numpy.zeros(len(metrics))

    if metrics.size == scores.size:
        return [metrics, scores]
    else:
        raise SystemExit("ERROR: unable to calculate accuracy metrics.")


def get_nbins_histogram(data):
    """
    Calculating the number of bins and the width of those bins for a histogram.

    :param data: 1-d numpy array.
    :return: (n_bins, bin_width) n_bins: int for the number of bins. bin_width: float with the width of the bins.

    """
    n = data.shape[0]
    lq = numpy.percentile(data, 25)
    uq = numpy.percentile(data, 75)
    iqr = uq - lq
    bin_width = 2 * iqr * n ** (-1 / 3)
    n_bins = int((numpy.max(data) - numpy.min(data)) / bin_width) + 2
    return n_bins, float(bin_width)


def get_bin_centres(bin_edges, geometric=False):
    """
    A function to calculate the centre points of bins from the bin edges from a histogram
    e.g., numpy.histogram. My default the arithmetic mean is provided (max+min)/2 but the
    geometric mean can also be calculated sqrt(min*max), this is useful for logarithmically
    spaced bins.

    :param bin_edges: numpy array of the bin edges
    :param geometric: boolean, if False (default) then the arithmetic mean return if True
                      then the geometric mean is returned.
    :returns: bin_centres - numpy array

    """
    if geometric:
        bin_centres = numpy.sqrt(bin_edges[1:] * bin_edges[:-1])
    else:
        bin_centres = (bin_edges[1:] + bin_edges[:-1]) / 2
    return bin_centres


def calc_otsu_threshold(data):
    """
    A function to calculate otsu's threshold for a dataset. Input is expected
    to be a 1d numpy array.

    Wikipedia, https://en.wikipedia.org/wiki/Otsu's_Method

    :param data: 1d numeric numpy array
    :returns: float (threshold)

    """
    # Note, this is based on the implementation within scikit-image

    # Calculate the histogram
    n_bins, bin_width = get_nbins_histogram(data)
    hist, bin_edges = numpy.histogram(data, bins=n_bins)
    bin_centres = get_bin_centres(bin_edges)
    # print(bin_centres)

    # Normalization so we have probabilities-like values (sum=1)
    hist = hist.astype(numpy.float32)
    hist = 1.0 * hist / numpy.sum(hist)

    # class probabilities for all possible thresholds
    weight1 = numpy.cumsum(hist)
    weight2 = numpy.cumsum(hist[::-1])[::-1]
    # class means for all possible thresholds
    mean1 = numpy.cumsum(hist * bin_centres) / weight1
    mean2 = (numpy.cumsum((hist * bin_centres)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of ``weight1``/``mean1`` should pair with zero values in
    # ``weight2``/``mean2``, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    return bin_centres[numpy.argmax(variance12)]


def calc_yen_threshold(data):
    """
    A function to calculate yen threshold for a dataset. Input is expected
    to be a 1d numpy array.

    Yen J.C., Chang F.J., and Chang S. (1995) "A New Criterion
    for Automatic Multilevel Thresholding" IEEE Trans. on Image
    Processing, 4(3): 370-378. DOI:`10.1109/83.366472`

    :param data: 1d numeric numpy array
    :returns: float (threshold)

    """
    # Note, this is based on the implementation within scikit-image

    # Calculate the histogram
    n_bins, bin_width = get_nbins_histogram(data)
    hist, bin_edges = numpy.histogram(data, bins=n_bins)
    bin_centres = get_bin_centres(bin_edges)

    # Normalization so we have probabilities-like values (sum=1)
    hist = hist.astype(numpy.float32)
    hist = 1.0 * hist / numpy.sum(hist)

    # Calculate probability mass function
    pmf = hist.astype(numpy.float32) / hist.sum()
    p1 = numpy.cumsum(pmf)  # Cumulative normalized histogram
    p1_sq = numpy.cumsum(pmf**2)
    # Get cumsum calculated from end of squared array:
    p2_sq = numpy.cumsum(pmf[::-1] ** 2)[::-1]
    # P2_sq indexes is shifted +1. I assume, with P1[:-1] it's help avoid
    # '-inf' in crit. ImageJ Yen implementation replaces those values by zero.
    crit = numpy.log(
        ((p1_sq[:-1] * p2_sq[1:]) ** -1) * (p1[:-1] * (1.0 - p1[:-1])) ** 2
    )
    return bin_centres[crit.argmax()]


def calc_isodata_threshold(data):
    """
    A function to calculate inter-means threshold for a dataset. Input is expected
    to be a 1d numpy array. Histogram-based threshold, known as Ridler-Calvard method
    or inter-means. Threshold values returned satisfy the following equality::

    threshold = (data[data <= threshold].mean() + data[data > threshold].mean()) / 2.0

    Ridler, TW & Calvard, S (1978), "Picture thresholding using an iterative selection
    method" IEEE Transactions on Systems, Man and Cybernetics 8: 630-632,
    DOI:`10.1109/TSMC.1978.4310039`

    :param data: 1d numeric numpy array
    :return: float (threshold)

    """
    # Note, this is based on the implementation within scikit-image

    # Calculate the histogram
    n_bins, bin_width = get_nbins_histogram(data)
    hist, bin_edges = numpy.histogram(data, bins=n_bins)
    bin_centres = get_bin_centres(bin_edges)

    # Normalization so we have probabilities-like values (sum=1)
    hist = hist.astype(numpy.float32)
    hist = 1.0 * hist / numpy.sum(hist)

    # csuml and csumh contain the count of pixels in that bin or lower, and
    # in all bins strictly higher than that bin, respectively
    csuml = numpy.cumsum(hist)
    csumh = csuml[-1] - csuml

    # intensity_sum contains the total pixel intensity from each bin
    intensity_sum = hist * bin_centres

    # l and h contain average value of all pixels in that bin or lower, and
    # in all bins strictly higher than that bin, respectively.
    # Note that since exp.histogram does not include empty bins at the low or
    # high end of the range, csuml and csumh are strictly > 0, except in the
    # last bin of csumh, which is zero by construction.
    # So no worries about division by zero in the following lines, except
    # for the last bin, but we can ignore that because no valid threshold
    # can be in the top bin.
    # To avoid the division by zero, we simply skip over the last element in
    # all future computation.
    csum_intensity = numpy.cumsum(intensity_sum)
    lower = csum_intensity[:-1] / csuml[:-1]
    higher = (csum_intensity[-1] - csum_intensity[:-1]) / csumh[:-1]

    # isodata finds threshold values that meet the criterion t = (l + m)/2
    # where l is the mean of all pixels <= t and h is the mean of all pixels
    # > t, as calculated above. So we are looking for places where
    # (l + m) / 2 equals the intensity value for which those l and m figures
    # were calculated -- which is, of course, the histogram bin centers.
    # We only require this equality to be within the precision of the bin
    # width, of course.
    all_mean = (lower + higher) / 2.0
    bin_width = bin_centres[1] - bin_centres[0]

    # Look only at thresholds that are below the actual all_mean value,
    # for consistency with the threshold being included in the lower pixel
    # group. Otherwise can get thresholds that are not actually fixed-points
    # of the isodata algorithm. For float images, this matters less, since
    # there really can't be any guarantees anymore anyway.
    distances = all_mean - bin_centres[:-1]
    thresholds = bin_centres[:-1][(distances >= 0) & (distances < bin_width)]

    return thresholds[0]


def calc_hist_cross_entropy(data, threshold):
    """
    A function which computes the cross-entropy between distributions
    above and below a threshold. Cross-entropy is a measure of the difference
    between two probability distributions for a given random variable or set of events.

    See Li and Lee (1993); this is the objective function ``threshold_li``
    minimizes. This function can be improved but this implementation most
    closely matches equation 8 in Li and Lee (1993) and equations 1-3 in
    Li and Tam (1998).

    Li C.H. and Lee C.K. (1993) "Minimum Cross Entropy Thresholding"
    Pattern Recognition, 26(4): 617-625 DOI:`10.1016/0031-3203(93)90115-D`

    Li C.H. and Tam P.K.S. (1998) "An Iterative Algorithm for Minimum
    Cross Entropy Thresholding" Pattern Recognition Letters, 18(8): 771-776
    DOI: `10.1016/S0167-8655(98)00057-9`

    :param data: 1d numeric numpy array
    :param threshold: float spliting to the two parts of the histogram.
    :returns: float (cross-entropy target value)

    """
    # Note, this is based on the implementation within scikit-image

    n_bins, bin_width = get_nbins_histogram(data)
    histogram, bin_edges = numpy.histogram(data, bins=n_bins, density=True)
    bin_centres = get_bin_centres(bin_edges)

    # Make sure values are positive and not zero.
    if numpy.min(bin_centres) < 1:
        offset = abs(numpy.min(bin_centres)) + 1
        bin_centres = bin_centres + offset
        threshold = threshold + offset

    t = numpy.flatnonzero(bin_centres > threshold)[0]
    m0a = numpy.sum(histogram[:t])  # 0th moment, background
    m0b = numpy.sum(histogram[t:])
    m1a = numpy.sum(histogram[:t] * bin_centres[:t])  # 1st moment, background
    m1b = numpy.sum(histogram[t:] * bin_centres[t:])
    mua = m1a / m0a  # mean value, background
    mub = m1b / m0b

    nu = -m1a * numpy.log(mua) - m1b * numpy.log(mub)
    return nu


def calc_li_threshold(data, tolerance=None, initial_guess=None):
    """
    A function which calculates a threshold value by Li's iterative
    Minimum Cross Entropy method.

    Li C.H. and Lee C.K. (1993) "Minimum Cross Entropy Thresholding"
    Pattern Recognition, 26(4): 617-625 DOI:`10.1016/0031-3203(93)90115-D`


    :param data: 1d numeric numpy array
    :param tolerance: float (optional) - Finish the computation when the
                     change in the threshold in an iteration is less than
                     this value. By default, this is half the smallest
                     difference between data values.
    :param initial_guess: float (optional) - Li's iterative method uses
                          gradient descent to find the optimal threshold.
                          If the histogram contains more than two modes
                          (peaks), the gradient descent could get stuck
                          in a local optimum. An initial guess for the
                          iteration can help the algorithm find the
                          globally-optimal threshold.
    :returns: float (threshold)

    """
    # Note, this is based on the implementation within scikit-image

    # At this point, the data only contains numpy.inf, -numpy.inf, or valid numbers
    data = data[numpy.isfinite(data)]

    # Li's algorithm requires positive image (because of log(mean))
    offset = 0.0
    if numpy.min(data) < 1:
        offset = abs(numpy.min(data)) + 1
        data = data + offset
        if initial_guess is not None:
            initial_guess = initial_guess + offset

    # If not provided define the tolerance
    if tolerance is None:
        tolerance = numpy.min(numpy.diff(numpy.unique(data))) / 2

    # Initial estimate for iteration. See "initial_guess" in the parameter list
    if initial_guess is None:
        t_next = numpy.mean(data)
    else:
        t_next = initial_guess

    # initial value for t_curr must be different from t_next by at
    # least the tolerance. Since the image is positive, we ensure this
    # by setting to a large-enough negative number
    t_curr = -2 * tolerance

    # Stop the iterations when the difference between the
    # new and old threshold values is less than the tolerance
    while abs(t_next - t_curr) > tolerance:
        t_curr = t_next
        foreground = data > t_curr
        mean_fore = numpy.mean(data[foreground])
        mean_back = numpy.mean(data[~foreground])

        t_next = (mean_back - mean_fore) / (numpy.log(mean_back) - numpy.log(mean_fore))

    threshold = t_next - offset
    return threshold


def calc_kurt_skew_threshold(
    data,
    max_val,
    min_val,
    init_thres,
    low_thres=True,
    contamination=10.0,
    only_kurtosis=False,
):
    """
    A function to calculate a threshold either side of the histogram based on

    :param data: 1d numeric numpy array
    :param max_val:
    :param min_val:
    :param init_thres:
    :param low_thres:
    :param contamination:
    :param only_kurtosis:
    :return:

    """
    import scipy.optimize
    import scipy.stats

    if len(data.shape) > 1:
        raise rsgislib.RSGISPyException("Expecting a single variable.")

    if (contamination < 1) or (contamination > 100):
        raise rsgislib.RSGISPyException(
            "contamination parameter should have a value between 1 and 100."
        )

    if low_thres:
        low_percent = numpy.percentile(data, contamination)
        if low_percent < max_val:
            max_val = low_percent

        if min_val >= max_val:
            min_val = numpy.min(data)
    else:
        up_percent = numpy.percentile(data, 100 - contamination)
        if up_percent > min_val:
            min_val = up_percent

        if max_val <= min_val:
            max_val = numpy.max(data)

    if min_val == max_val:
        print("Min: {}".format(min_val))
        print("Max: {}".format(max_val))
        raise rsgislib.RSGISPyException("Min and Max values are the same.")
    elif min_val > max_val:
        print("Min: {}".format(min_val))
        print("Max: {}".format(max_val))
        raise rsgislib.RSGISPyException(
            "Min value is greater than max - note this can happened if the "
            "contamination parameter caused threshold to be changed."
        )

    if (init_thres < min_val) or (init_thres > max_val):
        init_thres = min_val + ((max_val - min_val) / 2)

    def _opt_fun(x, *args):
        data = args[0]
        if low_thres:
            # Subset by x threshold
            data_sub = data[data > x]
        else:
            # Subset by x threshold
            data_sub = data[data < x]

        # Calculate kurtosis and skewness
        kurtosis = scipy.stats.kurtosis(data_sub)
        if only_kurtosis:
            kur_skew = kurtosis
        else:
            skew = scipy.stats.skew(data_sub)
            # Product of kurtosis and skewness
            kur_skew = abs(kurtosis) + abs(skew)

        return kur_skew

    opt_rslt = scipy.optimize.dual_annealing(
        _opt_fun, bounds=[(min_val, max_val)], args=[data], x0=[init_thres]
    )

    out_thres = None
    if opt_rslt.success:
        out_thres = opt_rslt.x[0]
    else:
        raise rsgislib.RSGISPyException("Optimisation failed, no threshold found.")

    return out_thres


def mask_data_to_valid(
    data: numpy.array, lower_limit: float = None, upper_limit: float = None
):
    """
    A function which removes rows from an nxm numpy array which are
    not finite or outside of the upper and lower thresholds provided.

    :param data: nxm numpy array (n rows, m features)
    :param lower_limit: lower threshold valid data is greater than this value.
    :param upper_limit: upper threshold valid data is lower than this value.

    """
    data = data[numpy.all(numpy.isfinite(data), axis=1)]
    if lower_limit is not None:
        data = data[numpy.all(data > lower_limit, axis=1)]
    if upper_limit is not None:
        data = data[numpy.all(data < upper_limit, axis=1)]
    return data


def bias_score(y_true, y_pred):
    """
    A function to calculate the regression model bias for each response variable
    in a multivariate regression.

    :param y_true: numpy.ndarray of true y values, shape=(n_samples, n_responses).
    :param y_true: numpy.ndarray of predicted y values, shape=(n_samples, n_responses).
    :return: bias (absolute bias), norm_bias (normalised bias (aka percentage bias))

    """
    import numpy

    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.array(y_pred)
    if y_true.ndim != y_pred.ndim:
        raise rsgislib.RSGISPyException("Error: y_true dimensions != y_pred.")
    if y_true.size != y_pred.size:
        raise rsgislib.RSGISPyException("Error: y_true size != y_pred.")

    bias = numpy.mean(y_pred - y_true, axis=0)
    norm_bias = (bias / numpy.mean(y_true, axis=0)) * 100
    return bias, norm_bias


def decompose_bias_variance(y_true, y_pred):
    """
    A function to perform bias-variance decomposition.
    Decompose the mean squared error into: Bias^2 + Variance + Irreducible Error.
    Useful example: https://towardsdatascience.com/mse-and-bias-variance-decomposition-77449dd2ff55

    :param y_true: numpy.ndarray of true y values, shape=(n_samples, n_responses).
    :param y_true: numpy.ndarray of predicted y values, shape=(n_samples, n_responses).
    :return: set (mse, bias_squared, variance, noise)

    """
    import numpy

    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.array(y_pred)
    if y_true.ndim != y_pred.ndim:
        raise SystemExit("Error: y_true dimensions != y_pred.")
    if y_true.size != y_pred.size:
        raise SystemExit("Error: y_true size != y_pred.")

    mse = numpy.mean((y_pred - y_true) ** 2, axis=0)
    bias_squared = numpy.mean(y_pred - y_true, axis=0) ** 2
    variance = numpy.var(y_pred - y_true, axis=0)
    noise = mse - (bias_squared + variance)
    if isinstance(noise, numpy.ndarray):
        noise[noise < 0] = 0
    else:
        if noise < 0:
            noise = 0
    return mse, bias_squared, variance, noise
